/* Copyright (C) 2017 Mono Wireless Inc. All Rights Reserved.    *
 * Released under MW-SLA-*J,*E (MONO WIRELESS SOFTWARE LICENSE   *
 * AGREEMENT).                                                   */

/*
 * 本コードは超簡単！TWELITE標準アプリのUARTメッセージの解釈、及び
 * 各種コマンドの送信を行うサンプルです。
 */

#include <SoftwareSerial.h>
#include <sercmd.h>         // 書式形式のパーサー

String VersionNumber = "1.0.0";

// typedef
typedef unsigned short uint16;
typedef unsigned long uint32;

// 受信したデータを保持する構造体
typedef struct
{
  byte LID;               // 送信元の論理デバイスID
  byte CMD;               // コマンド
  uint32 SID;      // 送信元のシリアルナンバー
//  byte IDE;               // パケット識別子
//  byte PVE;               // プロトコルバージョン
  byte LQI;               // LQI
  byte RID;               // 受信側の論理デバイスID
  byte RNM;               // 中継回数
  float TMS;              // タイムスタンプ
  uint16 PWR;     // 電源電圧(mV)
  byte DI;                // DIの状態 (1だったらLo)
//  byte DIM;               // DIのマスク(1度でも変化したら1)
  uint16 AD[4];   // アナログ電圧
} tsAppTwelite;

// Defines
#define S_OCTET(c) { *q = c; q++; }
#define S_BE_WORD(c) { \
  *q = ((c)>>8)&0xff; q++; \
  *q = ((c)&0xff); q++; \
  }
#define S_BE_DWORD(c) { \
  *q = ((c)>>24)&0xff; q++; \
  *q = ((c)>>16)&0xff; q++; \
  *q = ((c)>>8)&0xff; q++; \
  *q = ((c)&0xff); q++; \
  }

#define SW 8
#define LED 9

// Prototype
void vPutChar( unsigned char chr );    // 1バイトシリアル出力する関数を登録する。
bool bParseAppTwelite( sercmd ser, tsAppTwelite* sAppTwelite );   // 読み込んだバイト列をアナログ値などに変換する関数
void vShowData(tsAppTwelite sAppTwelite);                   // 変換したデータをPCに出力する関数
bool bTWELITEDataArrived();   // TWELITEからシリアルデータが来ているかどうか
void vSend80Commond( byte Addr, byte DO, uint16* PWMDuty );            // 0x80コマンドを送信する

// Global Object
SoftwareSerial MWSerial(2, 3);                   // RX, TX
sercmd SerialIn(false, 0, NULL );            // シリアル読み込み用のオブジェクト
sercmd SerialOut(false, 0, vPutChar );      // シリアル書き込み用のオブジェクト

void setup() {
  // シリアルの初期化
  Serial.begin(38400);

  // 起動メッセージの出力
  Serial.print("MONOWIRELESS App_Twelite Reader v");
  Serial.println(VersionNumber);
  Serial.println();

  pinMode( SW, INPUT );
  pinMode( LED, OUTPUT );
  digitalWrite( LED, LOW );

  // デフォルトの115200だと取りこぼすので38400に設定
  MWSerial.begin(38400);
}

void loop() {
  /*
   * TWELITEからのシリアルデータを読み込み、PCへシリアル出力する
   */
  if(bTWELITEDataArrived()){
      // 0x81コマンドだったら表示
      tsAppTwelite sAppTwelite;
      if(bParseAppTwelite(SerialIn, &sAppTwelite)){
        vShowData(sAppTwelite);
        if(sAppTwelite.DI&0x01){
          digitalWrite(LED, true);
        }else if( !(sAppTwelite.DI&0x01) ){
          digitalWrite(LED, false);
        }
      }
  }

  // タクトスイッチの状態を確認し、前の状態と変化していたら送信する
  static byte SwState = 0x00;
  if ( SwState != digitalRead(SW) ) {
    SwState = !SwState;
    // 無線送信をする
    uint16 PWM[] = {0,0,0,0};
    vSend80Commond( 0x78, SwState, PWM );
  }
}

// 書式形式出力用に1バイト出力関数
void vPutChar( unsigned char chr )
{
  MWSerial.write(chr); // 1バイトシリアル出力する関数
}

// TWELITEからシリアルデータが来ているか判定する。
bool bTWELITEDataArrived()
{
  // TWELITEからの出力を確認する。
  int c = MWSerial.read();
  if (c >= 0) {
    // １バイトずつ解釈をしていく
    byte u8stat = SerialIn.u8Parse(c);
    // 解釈に成功したら終了
    if (u8stat == E_SERCMD_COMPLETE) {
      return true;
    } else if (u8stat == E_SERCMD_CHECKSUM_ERROR) {
      // 電文中のチェックサムと計算したチェックサムが異なっているので、おそらく正しく読み込めなかった
      Serial.print("Command checksum error\n");
    }
  }
  return false;
}

// シリアルデータを解釈する 解釈出来たらTrue,出来なかったらFalseを返す
bool bParseAppTwelite( sercmd cmd, tsAppTwelite* sAppTwelite )
{
  sAppTwelite->LID = cmd.au8data[0];
  sAppTwelite->CMD = cmd.au8data[1];
  // コマンドIDが0x81かつバイト列が23バイトの時はIO変化のパケットなので解釈する
  if(sAppTwelite->CMD == 0x81 && cmd.u16len == 23){
//    sAppTwelite->IDE = cmd.au8data[2];
//    sAppTwelite->PVE = cmd.au8data[3];
    sAppTwelite->LQI = cmd.au8data[4];
    sAppTwelite->SID = ((unsigned long)cmd.au8data[5]<<24)|((unsigned long)cmd.au8data[6]<<16)|(cmd.au8data[7]<<8)|(cmd.au8data[8]);
    sAppTwelite->RID = cmd.au8data[9];
    sAppTwelite->TMS = (cmd.au8data[10]<<8|cmd.au8data[11])/64.0;
    sAppTwelite->RNM = cmd.au8data[12];
    sAppTwelite->PWR = cmd.au8data[13]<<8|cmd.au8data[14];
    sAppTwelite->DI = cmd.au8data[16];
//    sAppTwelite->DIM = cmd.au8data[17];
    // アナログ値の計算を行う
    for(int i=0; i<4; i++){
      sAppTwelite->AD[i] = cmd.au8data[18+i]<<2;
      sAppTwelite->AD[i] += ( cmd.au8data[22]>>(2*i) )&0x03;
      sAppTwelite->AD[i] = sAppTwelite->AD[i]<<2;
    }
    return true;
  }
  return false;
}

// 表示する
void vShowData(tsAppTwelite sAppTwelite)
{
  if(sAppTwelite.CMD == 0x81){
    Serial.print("Logical ID = 0x");Serial.println(sAppTwelite.LID, HEX);
    Serial.print("Command ID = 0x");Serial.println(sAppTwelite.CMD, HEX);
    Serial.print("LQI = ");Serial.println(sAppTwelite.LQI);
    Serial.print("Serial ID = 0x");Serial.println(sAppTwelite.SID, HEX);
//    Serial.print("Receive ID = 0x");Serial.println(sAppTwelite.RID, HEX);
    Serial.print("Timer = ");Serial.println(sAppTwelite.TMS);
    Serial.print("Number of Relay = ");Serial.println(sAppTwelite.RNM);
    Serial.print("Power Voltage = ");Serial.println(sAppTwelite.PWR);
    Serial.print("DI = ");
    // DIの状態ビット (普通に出力すると先頭の0が消えるため、1ビットずつ出力する)
    for(int i=3; i>=0; i-- ){
      Serial.print((sAppTwelite.DI&(1<<i))>>i, BIN);
    }
    Serial.println();

    // DIの変更状態ビット (普通に出力すると先頭の0が消えるため、1ビットずつ出力する)
/*    Serial.print("DI Mask = ");
    for(int i=3; i>=0; i-- ){
      Serial.print((sAppTwelite.DIM&(1<<i))>>i, BIN);
    }
    Serial.println();*/
    Serial.print("ADC1 = ");Serial.println(sAppTwelite.AD[0]);
    Serial.print("ADC2 = ");Serial.println(sAppTwelite.AD[1]);
    Serial.print("ADC3 = ");Serial.println(sAppTwelite.AD[2]);
    Serial.print("ADC4 = ");Serial.println(sAppTwelite.AD[3]);
    Serial.println();
  }  
}

void vSend80Commond( byte Addr, byte DO, uint16* PWMDuty )
{
  byte* q = SerialOut.au8data;

  S_OCTET(Addr);  // 送信先
  S_OCTET(0x80);  // 0x80コマンド
  S_OCTET(0x01);  // 1で固定
  S_OCTET(DO);    // 送信先のDOの状態を指定する
  S_OCTET(0x0F);  // IO状態設定マスク

  // PWMのDuty
  for(byte i = 0; i<4; i++){
    S_BE_WORD(PWMDuty[i]);
  }

  // バイト列の長さを指定する
  SerialOut.u16len = q - SerialOut.au8data;
  // TWELITEにコマンドを送信する
  SerialOut.vOutput();

  return;
}
